unit FUnionFind;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, StdCtrls;

type
  TForm1 = class(TForm)
    BRun: TButton;
    IMGInit: TImage;
    IMGResult: TImage;
    Memo: TMemo;
    Label1: TLabel;
    procedure BRunClick(Sender: TObject);

  Private
    Procedure TranfertImageToImageArray;
    Procedure TransfertCCRootArrayToDisplayImage;

    Procedure CCinit(W,H:Integer);
    Procedure CCclear;
    Function  CCfind(PosI:Integer):Integer;
    Function  CCunion(Root0,Root1:Integer):Integer;
    Function  CCadd(PosI,Root:Integer):Integer;
    Procedure CClabel(W,H:Integer);
  End;

Const
  CC_NOROOT = -1;

Var
  Image   :Array of Integer;
  CCroots :Array of Integer;

Var
  Form1: TForm1;

Implementation

{$R *.dfm}

//==============================================================================
Procedure TForm1.CCinit(W,H:Integer);
Begin
  // Array to store the input bitmap as [X,Y] integer values...
  SetLength(Image,   W*H);

  // Array to store the output result as [X,Y] integer labels numbers...
  SetLength(CCRoots, W*H);
End;

//==============================================================================
Procedure TForm1.CCclear;
Begin
  // In Delphi the array are destroyed automatically when the form is destroyed.
  // So, it's not usefull to free the array manualy.
End;

//==============================================================================
// Find the root of the node at position "PosI"
//==============================================================================
Function TForm1.CCfind(PosI:Integer):Integer;
Begin
  While (CCroots[PosI] <> PosI) do Begin
    PosI := CCroots[PosI];
  End;

  Result := PosI;
End;

//==============================================================================
// Union of the 2 path formed by the 2 roots
//==============================================================================
Function TForm1.CCunion(Root0,Root1:Integer):Integer;
Begin
  If      (Root0 = Root1)     Then Result := Root0
  Else If (Root0 = CC_NOROOT) Then Result := Root1
  Else If (Root1 = CC_NOROOT) Then Result := Root0
  Else If (Root0 < Root1)     Then Begin
    CCroots[Root1] := Root0;
    Result := Root0;
  End Else Begin
    CCroots[Root0] := Root1;
    Result := Root1;
  End;
End;

//==============================================================================
// Set the root of the node at position pos
//==============================================================================
Function TForm1.CCadd(PosI,Root:Integer):Integer;
Begin
  If (Root = CC_NOROOT) Then CCroots[PosI] := PosI
                        Else CCroots[PosI] := Root;
  Result := CCroots[PosI];
End;

//==============================================================================
// Return and build the connected component labels array
//==============================================================================
Procedure TForm1.CClabel(W,H:Integer);
Var x,y,root,PosI            :Integer;
    TheLabel                 :Integer;

Begin
  PosI := 0;

  For Y := 0 to H - 1 do Begin
    For X := 0 to W - 1 do Begin
      Root := CC_NOROOT;

      If ( (X>0)   And (Image[PosI - 1] = Image[PosI]))               Then Root := CCunion(CCfind(PosI - 1),     Root);
      If ( (X>0)   And (Y>0) And (Image[PosI - 1 - W] = Image[PosI])) Then Root := CCunion(CCfind(PosI - 1 - W), Root);
      If ( (Y>0)   And (Image[PosI - W] = Image[PosI]))               Then Root := CCunion(CCfind(PosI - W),     Root);
      If ( (X<W-1) And (Y>0) And (Image[PosI + 1 - W] = Image[PosI])) Then Root := CCunion(CCfind(PosI + 1 - W), Root);

      Root := CCadd(PosI, Root);

      Inc(PosI);
    End;
  End;

  // Remove indirection...
  For PosI := 0 to (W * H) - 1 do Begin
    CCroots[PosI] := CCfind(PosI);
  End;

  // Relabel...
  TheLabel := 1;
  For PosI := 0 to (W * H) - 1 do Begin
    If (CCroots[PosI] = PosI) Then Begin
      Memo.Lines.Add(IntToStr(TheLabel));
      CCroots[PosI] := TheLabel;
      Inc(TheLabel);
    End Else Begin
      CCroots[PosI] := CCroots[CCroots[PosI]];
    End;
  End;
End;

//==============================================================================
// Run the process and display result.
//==============================================================================
Procedure TForm1.BRunClick(Sender: TObject);
Begin
  Memo.Clear; // Clear Memo component that display Labels values
  CCinit(448,146);
  TranfertImageToImageArray;
  CClabel(448,146);
  TransfertCCRootArrayToDisplayImage;
  CCclear;
End;

//==============================================================================
// Transfert the bitmap to the "Image" array.
// Each value is stored as an integer (TColor).
//==============================================================================
Procedure TForm1.TranfertImageToImageArray;
Var X,Y,I                    :Integer;
    Val1                     :Integer;
Begin
  I := 0;
//  IMGInit.Picture.LoadFromFile('D:\Development\Personnel\UnionFind\Labelling11.bmp');
  For X := 0 to 448 - 1 do Begin
    For Y := 0 to 146 - 1 do Begin
      Val1 := IMGInit.Picture.Bitmap.Canvas.Pixels[X,Y];
      Image[I] := Val1;
      Inc(I);
    End;
  End;
End;

//==============================================================================
// Transfert the "CCRoot" array values to the "TCanva" image to display.
// Each label has a defined color.
//==============================================================================
Procedure TForm1.TransfertCCRootArrayToDisplayImage;
Var X,Y,I                    :Integer;
    Val1                     :Integer;
Begin
  I := 0;
  For X := 0 to 448 - 1 do Begin
    For Y := 0 to 146 - 1 do Begin
      Val1 := CCroots[I];
      RandSeed := Val1;
      IMGResult.Canvas.Pixels[X,Y] := Random($FFFFFF);
      Inc(I);
    End;
  End;
End;

//==============================================================================
End.
